/* 
 * This file is part of ideas.
 *
 *  ideas is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  ideas is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with ideas.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * (c) Keith Webster Johnston & Markos Charatzas 
 */
package ideas.filters;

import ideas.persistence.StaticUnitOfWorkBinder;
import ideas.persistence.UnitOfWork;
import ideas.persistence.UnitOfWorkBinder;
import ideas.persistence.UnitsOfWork;
import ideas.resource.IdeaResource;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

/**
 * Creates a new long {@link UnitOfWork} which contains the request under a 
 * single long transaction.
 * 
 * <p>
 * The {@link UnitOfWork} is done at the end of the request lifecycle
 * 
 * <p>
 * The unitOfWork is {@link UnitOfWorkBinder bound} to a specified type that 
 * is meant to take part in fulfilling the request   
 * 
 * @see UnitsOfWork#newLongUnitOfWork()
 * @author Markos Charatzas [@qnoid]
 *
 */
public class UnitOfWorkFilter implements Filter 
{
	private UnitsOfWork unitOfWorks;
	private UnitOfWorkBinder binder;

	/* (non-Javadoc)
	 * @see javax.servlet.Filter#init(javax.servlet.FilterConfig)
	 */
	@Override
	public void init(FilterConfig filterConfig) throws ServletException 
	{
		this.unitOfWorks = UnitsOfWork.newInstance();				
		this.binder = StaticUnitOfWorkBinder.newBinder(IdeaResource.class);
	}

	/* (non-Javadoc)
	 * @see javax.servlet.Filter#doFilter(javax.servlet.ServletRequest, javax.servlet.ServletResponse, javax.servlet.FilterChain)
	 */
	@Override
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException 
	{
		UnitOfWork unitOfWork = 
				this.unitOfWorks.newLongUnitOfWorkBuilder()
					.bind(this.binder)
					.build();

		try
		{						
			chain.doFilter(request, response);
		}
		finally
		{
			unitOfWork.done();
		}
		
	}

	/* (non-Javadoc)
	 * @see javax.servlet.Filter#destroy()
	 */
	@Override
	public void destroy() {
		// TODO Auto-generated method stub
		
	}
}