/* 
 * This file is part of ideas.
 *
 *  ideas is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  ideas is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with ideas.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * (c) Keith Webster Johnston & Markos Charatzas 
 */
package ideas.persistence;


import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.TypedQuery;

/**
 * Provides access to {@link AtomicOperation}s
 * 
 * @see UnitOfWork#make(AtomicOperation)
 * 
 * @author Markos Charatzas [@qnoid]
 *
 */
public final class AtomicOperations 
{
	/**
	 * 
	 */
	private static final AtomicOperations SINGLETON = new AtomicOperations();

	/**
	 * @return singleton of <b>this</b>
	 */
	public static AtomicOperations singleton() {
	return SINGLETON;
	}
	
	/**
	 * Creates a new {@link AtomicOperation} to read all elements of the query 
	 * type specified
	 * 
	 * @param <E> the type of the elements returned by the named query
	 * @param queryType the type of the named query
	 * @return a new operation which will execute the specified named query
	 * and return the result
	 * @see NamedQueryTypes
	 * @see EntityManager#createNamedQuery(String, Class)
	 */
	public <E> AtomicOperation<List<E>> newReadOperation(final NamedQueryType<E> queryType) 
	{
	return new AtomicOperation<List<E>>() {

		@Override
		public List<E> make(EntityManager entityManager) 
		{
		  TypedQuery<E> namedQuery = 
		      entityManager.createNamedQuery( queryType.name(), queryType.type() );
		  
    return namedQuery.getResultList();				
		}
	};
	}

	/**
   * Creates a new {@link AtomicOperation} to create an entity
   *   
	 * @precondition the instance must not exist
	 * 
	 * @param instance the instance to create
	 * @return a new operation which will create the given instance.
	 * @see EntityManager#persist(Object)
	 */
	public <E> AtomicOperation<Void> newCreateOperation(final E instance) 
	{
	return new AtomicOperation<Void>() {

			@Override
			public Void make(EntityManager entityManager) {
					entityManager.persist(instance);
			return null;
			}
		};
		}

	/**
	 * Creates a new {@link AtomicOperation} to read an entity by its key  
	 * 
	 * @param <E> the type of the entity
	 * @param key the key holding the type of the entity and its id value
	 * @return a new {@link AtomicOperation} that finds an entity by its id 
	 * @see EntityManager#find(Class, Object)
	 */
	public <E> AtomicOperation<E> newReadOperationOnPrimaryKey(final Key<E> key) 
	{
	return new AtomicOperation<E>() {

			@Override
			public E make(EntityManager entityManager) {
			return entityManager.find(key.type(), key.value());
			}
		};
	}

	/**
	 * Creates a new {@link AtomicOperation} to create all elements provided 
	 * by the resource specified.
	 * 
	 * @param resource the resource holding the elements
	 * @return a new operation which will create all elements as returned
	 * by the resource
	 * @see EntityManager#persist(Object)
	 */
	public <E> AtomicOperation<Void> newCreateOperationOnListResource(final ListResource<E> resource) 
	{
	return new AtomicOperation<Void>() {

			@Override
			public Void make(EntityManager entityManager) 
			{
				List<E> elements = resource.get();
				
				for (E t : elements) {
					entityManager.persist(t);
				}
				
			return null;	
			}
		};
		}
}
