/* 
 * This file is part of ideas.
 *
 *  ideas is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  ideas is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with ideas.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * (c) Keith Webster Johnston & Markos Charatzas 
 */
package ideas.persistence;

import ideas.annotations.Cost;
import ideas.annotations.CostValue;

import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;

import net.jcip.annotations.NotThreadSafe;

/**
 * A {@link AtomicUnitOfWork} that can be used to execute a single 
 * {@link AtomicOperation} atomically.
 * 
 * <p>
 * {@link AtomicUnitOfWork} have no visibility among their 
 * {@link AtomicOperation} due to transaction isolation level.
 * 
 * <p>
 * A unit of work must be marked as {@link #done()} before discarded.
 * 
 * @see UnitsOfWork#newLongUnitOfWork()
 * @author Markos Charatzas [@qnoid]
 *
 */
@NotThreadSafe
@Cost(CostValue.INEXPENSIVE)
public class AtomicUnitOfWork implements UnitOfWork 
{
	/*
	 * 
	 */
	private final EntityManager entityManager;
	private final EntityTransaction transaction;

	private UnitOfWorkState state = UnitOfWorkState.NOT_DONE;
	
	/**
	 * @param entityManager
	 * @param transaction
	 */
	AtomicUnitOfWork(EntityManager entityManager, EntityTransaction transaction) 
	{
		this.entityManager = entityManager;
		this.transaction = transaction;
	}


	/**
	 * Makes the atomic operation inside a single transaction.
	 * <p>
	 * The operation is guaranteed to commit or rollback in case of an exception
	 *
   * @throws UnitOfWorkException if closed
   * @see UnitOfWork#make(AtomicOperation)
	 */
	@Override
	public <Result> Result make(AtomicOperation<Result> operation) 
	{
		UnitOfWorkState.throwExceptionIfDone(this.state);
		
		try
		{
			this.transaction.begin();

		return operation.make(this.entityManager);
		}
		catch (Exception e) 
		{
			this.transaction.rollback();
		throw new UnitOfWorkException(e);
		}
		finally{    
			this.transaction.commit();
		}
	}

	/**
	 * Closes the underlying {@link EntityManager}
	 * 
	 * @throws UnitOfWorkException if closed
	 * @see UnitOfWork#done()
	 */
	@Override
	public void done() 
	{
		UnitOfWorkState.throwExceptionIfDone(this.state);
		
		this.entityManager.close();
		this.state = UnitOfWorkState.DONE;
	}
}
