/* 
 * This file is part of ideas.
 *
 *  ideas is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  ideas is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with ideas.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * (c) Keith Webster Johnston & Markos Charatzas 
 */
package ideas.persistence;

import ideas.annotations.UnitOfWorkType;

import java.lang.reflect.Field;

/**
 * Binds a {@link UnitOfWork} to a static field declared in a specified type
 * 
 * @author Markos Charatzas [@qnoid]
 *
 */
public final class StaticUnitOfWorkBinder implements UnitOfWorkBinder 
{
	/**
   * 
   */
  private static final String IS_FIELD_ANNOTATED_MESSAGE = String.format(
      "No annotated %s field found", 
      UnitOfWorkType.class.getCanonicalName());

  /**
	 * @param type the type declaring the annotated field
	 * @return a new {@link UnitOfWorkBinder}
	 */
	public static StaticUnitOfWorkBinder newBinder(Class<?> type) {
	return new StaticUnitOfWorkBinder(type);
	}

	private final Class<?> type;

	private StaticUnitOfWorkBinder(Class<?> type) {
		this.type = type;
	}

	/**
	 * Binds the specified {@link UnitOfWork unitOfWork} to a static field 
	 * declared in the type as specified in the constructor.
	 *
	 * @precondition the field must be declared in the type
	 * @precondition the field must be annotated by {@link UnitOfWorkType}
	 * @precondition the field must be static
	 * 
	 * @postcondition the unitOfWork will be set as the value of the field
	 * 
	 * @param unitOfWork the unit of work to set
	 * @see ideas.persistence.UnitOfWorkBinder#bind(U)
	 * @see Class#getDeclaredFields()
	 * @throws UnitOfWorkTypeException if field is not annotated or non static
	 */
	@Override
	public <U extends UnitOfWork> void bind(U unitOfWork) 
	{
		Field[] declaredFields = this.type.getDeclaredFields();
		
		for (Field field : declaredFields) 
		{
			if(field.isAnnotationPresent(UnitOfWorkType.class))
			{
				try {
					field.setAccessible(true);
					field.set(null, unitOfWork);
				return;
				} 
				catch (IllegalArgumentException e) {
					throw new RuntimeException(e);
				} 
				catch (IllegalAccessException e) {
					throw new RuntimeException(e);
				}
				catch(NullPointerException e) {
					throw new UnitOfWorkTypeException("Is the annotated @UnitOfWorkType static?");
				}
			}
		}
		
	throw new UnitOfWorkTypeException(IS_FIELD_ANNOTATED_MESSAGE);
	}
}
