/* 
 * This file is part of ideas.
 *
 *  ideas is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  ideas is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with ideas.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * (c) Keith Webster Johnston & Markos Charatzas 
 */
package ideas.persistence;

import java.util.Arrays;
import java.util.List;

/**
 * Provides a list of methods for all possible {@link AtomicOperations}.
 * 
 * <p>
 * The decorator is not guaranteed to be maintained and the 
 * {@link AtomicOperations} class should be used as a reference for any
 * {@link AtomicOperation} available to {@link #make(AtomicOperation)} 
 * 
 * @author Markos Charatzas [@qnoid]
 *
 */
public class UnitOfWorkDecorator implements UnitOfWork 
{
	private final AtomicOperations operations;
	private final UnitOfWork unitOfWork;

	UnitOfWorkDecorator(AtomicOperations operations, UnitOfWork unitOfWork) {
		this.operations = operations;
		this.unitOfWork = unitOfWork;
	}

	/**
	 * Lists all the elements as specified by the {@link NamedQueryType}
	 * 
	 * @param namedQueryType
	 * @return all elements based on the query type
	 * @see NamedQueryTypes
	 * @see AtomicOperations#newReadOperation(NamedQueryType)
	 */
	public final <E> List<E> list(NamedQueryType<E> namedQueryType) {
		AtomicOperation<List<E>> newReadOperation = 
				this.operations.newReadOperation(namedQueryType);
		
	return this.make(newReadOperation);
	}

	/**
	 * Creates the instance
	 * 
	 * @param <E> the type of the instance
	 * @param instance the instance to create
   * @see AtomicOperations#newCreateOperation(Object)
	 */
	public final <E> void create(E instance) {
		AtomicOperation<Void> newCreateOperation = 
				this.operations.newCreateOperation(instance);
				
		this.make(newCreateOperation);			
	}

	/**
	 * Creates all the elements
	 * 
   * @param <E> the type of the elements
   * @param elements the elements to create
   * @see AtomicOperations#newCreateOperationOnListResource(Object)
	 */
	public final <E> void createFromVarrargs(final E... elements) 
	{
		this.createfromListResource(new ListResource<E>() {
	
			@Override
			public List<E> get() {
			return Arrays.asList(elements);
			}
		}
		);
	}

	/**
	 * Creates all the elements provides by the resource
	 * 
	 * @param <E> the type of the elements
	 * @param listResource the resource providing the List of elements to create
	 * @see AtomicOperations#newCreateOperationOnListResource(ListResource)
	 */
	public final <E> void createfromListResource(ListResource<E> listResource) {
		AtomicOperation<Void> newCreateOperation = 
				this.operations.newCreateOperationOnListResource(listResource);
		
		this.make(newCreateOperation);
	}

	/**
	 * Retrieves an element by its primary key
	 * 
	 * @param <E> the element type
	 * @param key the primary key of the element
	 * @return the element identified by the key
	 * @see AtomicOperations#newReadOperationOnPrimaryKey(Key)
	 */
	public final <E> E read(Key<E> key) {
		AtomicOperation<E> newReadOperation = 
				this.operations.newReadOperationOnPrimaryKey(key);
		
	return this.make(newReadOperation);
	}

  /* (non-Javadoc)
   * @see ideas.persistence.UnitOfWork#make(ideas.persistence.AtomicOperation)
   */
  @Override
  public <Result> Result make(AtomicOperation<Result> operation) {
  return this.unitOfWork.make(operation);
  }

  /* (non-Javadoc)
   * @see ideas.persistence.UnitOfWork#done()
   */
  @Override
  public void done() {
    this.unitOfWork.done();
  }
}