/* 
 * This file is part of ideas.
 *
 *  ideas is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  ideas is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with ideas.  If not, see <http://www.gnu.org/licenses/>.
 *  
 * (c) Keith Webster Johnston & Markos Charatzas 
 */
package ideas.persistence;

import ideas.annotations.Cost;
import ideas.annotations.CostValue;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityTransaction;
import javax.persistence.Persistence;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Provides access to {@link UnitOfWork units of work}
 * 
 * @author Markos Charatzas [@qnoid]
 *
 */
public final class UnitsOfWork 
{
  /**
	 * Use to build a single {@link LongUnitOfWork}.
	 * 
	 * Defaults:
	 *     binder = {@link UnitOfWorkBinders#NO_OP}
	 *     
	 * <p>
	 * The builder allows you to bind a {@link UnitOfWork}
	 */
	public final class LongUnitOfWorkBuilder
	{
		private final EntityManager entityManager;
		private UnitOfWorkBinder binder = UnitOfWorkBinders.NO_OP;
				
		LongUnitOfWorkBuilder(EntityManager entityManager) {
			this.entityManager = entityManager;
		}

		/**
		 * Sets a binder for the {@link UnitOfWork} built by the builder
		 * 
		 * @param binder
		 * @return <b>this</b> for chaining
		 */
		public <T> LongUnitOfWorkBuilder bind(UnitOfWorkBinder binder)
		{
			this.binder = binder;
		return this;
		}
		
		/**
		 * @return a new LongUnitOfWork
		 * @see UnitsOfWork#newLongUnitOfWork()
		 */
		/*
		 * @see UnitOfWorks#newLongAtomicUnitOfWork(EntityManager)
		 */
		public UnitOfWork build()
		{
		  UnitOfWork longAtomicUnitOfWork = 
			    UnitsOfWork.this.newLongAtomicUnitOfWork(this.entityManager);
					
			this.binder.bind(longAtomicUnitOfWork);
			
		return longAtomicUnitOfWork;
		}
	}
	/*
	 * Singleton Holder as described by Joshua Bloch, Effective Java 2nd edition,
	 * Item 71, p283
	 *
	 */
	private static final class EntityManagerFactoryHolder
	{
	  /**
	   * 
	   */
	  private static final String PERSISTENCE_UNIT_NAME = "persistenceUnit";

		/*
		 * <blockquote>
		 * A EntityManagerFactory is an expensive-to-create, threadsafe object 
		 * intended to be shared by all application threads. 
		 * 
		 * It is created once, usually on application startup.
		 * 
		 * </blockquote>
		 * Hibernate EntityManager
		 * User guide
		 * 
		 * 5.1. Entity manager and transaction scopes
		 * 
		 * http://docs.jboss.org/hibernate/stable/entitymanager/reference/en/html_single/
		 */
		final static EntityManagerFactory SINGLETON = 
				createEntityManagerFactory();
	
		private static EntityManagerFactory createEntityManagerFactory() 
		{
			try {
				return Persistence.createEntityManagerFactory( PERSISTENCE_UNIT_NAME );
			} 
			catch (Exception e) 
			{
				LOGGER.error(e.getMessage(), e);
				throw new RuntimeException("Error creating entity manager factory", e);
			}
		}
	}
	
	private static final Logger LOGGER = LoggerFactory.getLogger(UnitsOfWork.class);
	
	/**
	 * @return a new instance
	 */
	@Cost(CostValue.INEXPENSIVE)
	public static UnitsOfWork newInstance() {
	return UnitsOfWork.newInstance(EntityManagerFactoryHolder.SINGLETON);
	}

	public static UnitsOfWork newInstance(EntityManagerFactory entityManagerFactory) {
	return new UnitsOfWork(entityManagerFactory);
	}
	
	/*
	 * A ThreadLocal is used to maintain an association between an EntityManager
	 * and a thread. This is required in order to support a concurrent environment
	 * where multiple threads need to be assigned a new EntityManager yet each
	 * needs to refer to a single instance.
	 * 
   * <p>
   * The initial value for every ThreadLocal is a new EntityManager, from then 
   * on whenever the thread requests for an EntityManager is given the same 
   * instance.
   * 
	 * <p>
	 * After calling EntityManager#close on an EntityManager obtained by the 
	 * ThreadLocal, you must call ThreadLocal#set(null) in order for a new
	 * EntityManager to be created the next time the thread asks for one.
	 * 
	 * <p>
	 * Failing to do so will result in an EntityManager throwing 
	 * IllegalStateException for any of its operations.
	 * 
	 */
	private final ThreadLocal<EntityManager> entityManagers = 
			new ThreadLocal<EntityManager>(){

				@Override
				protected EntityManager initialValue() {
				return entityManagerFactory.createEntityManager();
				}

				@Override
				public EntityManager get() 
				{
					EntityManager entityManager = super.get();
				
					if(entityManager == null){
						return entityManagerFactory.createEntityManager();
					}
					
				return entityManager;
				}
		};
		
	private final EntityManagerFactory entityManagerFactory;
		

	/**
	 * @param entityManagerFactory
	 */
	private UnitsOfWork(EntityManagerFactory entityManagerFactory) {
		this.entityManagerFactory = entityManagerFactory;
	}
	
	/**
	 * Creates a new {@link LongUnitOfWork} which is bound to a single 
	 * transaction as obtained by the specified entityManager. 
	 * 
	 * <p>
	 * In case of an exception during {@link UnitOfWork#make(AtomicOperation)
	 * making an AtomicOperation} the associated {@link EntityTransaction 
	 * transaction} will be rollbacked.
	 * 
	 * <p>
	 * Calling {@link UnitOfWork#done()} will commit the associated transaction
	 * and close the specified entityManager.
	 * 
	 * @postcondition the associated transaction has started
	 * 
	 * @param entityManager the entity manager to obtain the transaction from 
	 * @return a new {@link UnitOfWork}
	 */
  UnitOfWork newLongAtomicUnitOfWork(final EntityManager entityManager)
  {
    final EntityTransaction transaction = entityManager.getTransaction();
    transaction.begin();

    LongUnitOfWork longUnitOfWork = 
      new LongUnitOfWork(entityManager)
      {
          @Override
          public <Result> Result make(AtomicOperation<Result> operation)
          {
            try {
              return super.make(operation);
            }
            catch (Exception e) 
            {
              LOGGER.error(e.getMessage(), e);
              transaction.rollback();
            throw new UnitOfWorkException(e);
            }
          }

          @Override
          public void done() 
          {
            transaction.commit();
            entityManager.close();                          
            UnitsOfWork.this.entityManagers.set(null);                          
          }
      };
                  
  return longUnitOfWork;
  }
  
	/**
	 * @return a new {@link LongUnitOfWorkBuilder}
	 */
	public <T> LongUnitOfWorkBuilder newLongUnitOfWorkBuilder()
	{
		EntityManager entityManager = this.entityManagers.get();
		
	return new LongUnitOfWorkBuilder(entityManager);
	}
	
	/**
	 * Creates a new {@link UnitOfWork} that can be used to 
	 * {@link UnitOfWork#make(AtomicOperation) make} a set of 
	 * {@link AtomicOperation}s across a single transaction.
	 * 
	 * <p>
	 * All {@link AtomicOperation}s made within the same {@link UnitOfWork} are
	 * contained in a single transaction, hence have visibility between them.
	 * 
	 * <p>
   * Calling {@link UnitOfWork#done()} will commit the associated transaction.
   * 
   * <p>
   * The {@link UnitOfWork} returned must be marked as 
   * {@link UnitOfWork#done() done} <b>before a new one is requested by the
   * same thread.</b>
   * 
   * <p>
   * Failing to do so, there is not guarantee as to the behavior of all the
   * {@link UnitOfWork} returned by this method for a single thread. 
   * 
	 * @return a new @NoThreadSafe {@link UnitOfWork} 
	 */
	/*
	 * The reason behind this condition is that since there is a one-to-one 
	 * Thread to {@link EntityManager} relationship, there is a one-to-many on 
	 * {@link UnitOfWork}.
	 * 
	 * Either a precondition should be forced on the method for a fail early 
	 * exception in case where a previous UnitOfWork is not done for the requested
	 * thread or block the call until so.
	 * 
	 * @see #newLongAtomicUnitOfWork(EntityManager)
	 * 
	 */
	public <T> UnitOfWork newLongUnitOfWork() {
	return this.newLongUnitOfWorkBuilder().build();
	}

	/**
	 * Creates a new {@link AtomicUnitOfWork}
	 * 
   * <p>
   * The {@link AtomicUnitOfWork} associated with the operation is guaranteed to 
   * be {@link AtomicUnitOfWork#done() done} by the end of the execution even if
   * an exception occurs.
	 *  
	 * @return a new {@link UnitOfWork} as to 
	 * {@link UnitOfWork#make(AtomicOperation) make} the operation
	 */
	public AtomicUnitOfWork newAtomicUnitOfWork() 
	{
		EntityManager entityManager = this.entityManagers.get();
		
		EntityTransaction transaction = 
				entityManager.getTransaction();
		
		final AtomicUnitOfWork atomic = 
				new AtomicUnitOfWork(entityManager, transaction);
		
	return new AtomicUnitOfWork(entityManager, transaction) {
		
		@Override
		public <Result> Result make(AtomicOperation<Result> operation) 
		{
			try {
			return atomic.make(operation);
			}
			finally
			{
				this.done();
			}
		}
		
		@Override
		public void done() {
			atomic.done();
		}
	};
	}

  /**
   * Decorates the specified unitOfWork allowing execution of a number of
   * {@link AtomicOperation}s 
   *  
   * @param unitOfWork the unitOfWork to decorate
   * @return a new {@link UnitOfWorkDecorator}
   */
  public UnitOfWorkDecorator decorate(UnitOfWork unitOfWork)
  {
    AtomicOperations operations = AtomicOperations.singleton();  

  return new UnitOfWorkDecorator(operations, unitOfWork);
  }	
}

